﻿using System;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using UniRx;
using UnityEngine;
using static Magify.MagifyManager;

namespace Magify
{
    public class AdvertiserService : IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(MagifyService.LogScope);

        private readonly MagifySettings _settings;
        private readonly ServicePrefs _prefs;
        private readonly CampaignsProvider _campaigns;
        private readonly INetworkStatusProvider _network;
        private readonly AdPreloader _adPreloader;

        private readonly CompositeDisposable _disposables = new();

        private readonly ReactiveProperty<int> _bannerDeactivationsCounter = new(0);
        private readonly ReactiveProperty<bool> _bannerRequired = new(false);
        private readonly ReactiveProperty<bool> _bannerActive = new(false);

        private readonly Subject<Unit> _onBeforeShowRewardVideo = new();
        private readonly Subject<Unit> _onAfterShowRewardVideo = new();

        private readonly Subject<Unit> _onBeforeShowInterstitialVideo = new();
        private readonly Subject<Unit> _onAfterShowInterstitialVideo = new();

        public IReactiveProperty<bool> AdPreloadEnabled => _adPreloader.IsEnabled;

        /// <summary>
        /// Determines if a banner is required by application
        /// </summary>
        public IReactiveProperty<bool> IsBannerRequired => _bannerRequired;

        /// <summary>
        /// Determines if a banner currently visible for player.
        /// It can be false even when <see cref="IsBannerRequired"/> is true because someone could request a temporary disable using <see cref="TemporaryDisableBanner"/>
        /// </summary>
        public IReadOnlyReactiveProperty<bool> IsBannerActive => _bannerActive;

        public IObservable<Unit> OnBeforeShowRewardVideo => _onBeforeShowRewardVideo;
        public IObservable<Unit> OnAfterShowRewardVideo => _onAfterShowRewardVideo;
        public IObservable<Unit> OnBeforeShowInterstitialVideo => _onBeforeShowInterstitialVideo;
        public IObservable<Unit> OnAfterShowInterstitialVideo => _onAfterShowInterstitialVideo;

        public IAdsMediator Mediator { get; private set; }

        internal AdvertiserService(
            [NotNull] MagifySettings settings,
            [NotNull] ServicePrefs prefs,
            [NotNull] CampaignsProvider campaigns,
            [NotNull] ProductsObtainer products,
            [NotNull] AdPreloader adPreloader,
            [NotNull] INetworkStatusProvider network)
        {
            _settings = settings;
            _prefs = prefs;
            _campaigns = campaigns;
            _adPreloader = adPreloader;
            _network = network;

            _network.Reachability
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => RefreshBanner())
                .AddTo(_disposables);

            _bannerRequired
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => RefreshBanner())
                .AddTo(_disposables);

            _bannerDeactivationsCounter
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => RefreshBanner())
                .AddTo(_disposables);

            products.OnProductObtained
                .Subscribe(_ => RefreshBanner())
                .AddTo(_disposables);
        }

        void IDisposable.Dispose()
        {
            DisposeCurrentMediator();
            _disposables?.Dispose();
        }

        public void SetAdsMediator([CanBeNull] IAdsMediator mediator)
        {
            DisposeCurrentMediator();
            Mediator = mediator;
            if (Mediator != null)
            {
                Mediator.OnBannerLoaded += BannerLoadedHandler;
                Mediator.OnBannerLoadFailed += BannerLoadFailedHandler;
                Mediator.OnBannerClicked += BannerClickedHandler;
                RefreshBanner();
            }
            _adPreloader.SetAdsMediator(mediator);
        }

        #region Banner

        private void BannerLoadedHandler(IAdsImpression impression)
        {
#if MAGIFY_3_OR_NEWER
            var campaignName = _campaigns.GetCampaign(_settings.BannerDefaultEvent)?.Name;
            if (campaignName == null)
            {
                _logger.LogWarning($"The banner was loaded, but there was no campaign on the default banner event ({_settings.BannerDefaultEvent})");
            }
            TrackAdsImpression(campaignName ?? string.Empty, impression);
#else
            TrackAdsImpression(CampaignType.Banner, impression);
#endif
        }

        private void BannerLoadFailedHandler(string reason)
        {
#if MAGIFY_3_OR_NEWER
            var campaignName = _campaigns.GetCampaign(_settings.BannerDefaultEvent)?.Name;
            if (campaignName == null)
            {
                _logger.LogWarning($"The banner was failed to show, but there was no campaign on the default banner event ({_settings.BannerDefaultEvent})");
            }
            TrackImpressionFailFor(campaignName ?? string.Empty, reason);
#else
            TrackImpressionFailFor(CampaignType.Banner, reason);
#endif
        }

        private void BannerClickedHandler(IAdsImpression obj)
        {
#if MAGIFY_3_OR_NEWER
            var campaignName = _campaigns.GetCampaign(_settings.BannerDefaultEvent)?.Name;
            if (campaignName == null)
            {
                _logger.LogWarning($"The banner was clicked, but there was no campaign on the default banner event ({_settings.BannerDefaultEvent})");
            }
            TrackAdsClickFor(campaignName ?? string.Empty);
#else
            TrackAdsClickFor(CampaignType.Banner);
#endif
        }

        public IDisposable TemporaryDisableBanner()
        {
            return new DisposableCounter(_bannerDeactivationsCounter);
        }

        /// <summary>
        /// Refreshes current banner state
        /// </summary>
        public void RefreshBanner()
        {
            if (Mediator == null)
            {
                return;
            }

            // Check global state, current temporary deactivations and magify campaign availability to decide whether to show or hide the banner
            var needBanner = IsBannerRequired.Value && _bannerDeactivationsCounter.Value <= 0 && _campaigns.GetCampaignToHandle(_settings.BannerDefaultEvent, null, true) is BannerCampaign;

            _logger.Log($"{nameof(RefreshBanner)} - " +
                       $"GloballyRequired={IsBannerRequired.Value}; " +
                       $"TemporaryDeactivations={_bannerDeactivationsCounter.Value}; " +
                       $"ShouldBeShown={needBanner}; " +
                       $"CurrentlyShown={_bannerActive.Value};");

            // Do nothing if current banner state same as required 
            if (needBanner == _bannerActive.Value && needBanner == Mediator.IsBannerVisible)
            {
                _logger.Log($"{nameof(RefreshBanner)} - banner already {(needBanner ? "shown" : "hidden")}. ");
                return;
            }

            if (needBanner)
            {
                _campaigns.GetCampaignToHandle(_settings.BannerDefaultEvent, null);
                if (!_network.IsNetworkReachable)
                {
                    _logger.Log($"{nameof(RefreshBanner)} - can't show because there is no internet");
                    return;
                }
                _logger.Log($"{nameof(RefreshBanner)} - asking {nameof(IAdsMediator)} to show banner");
                Mediator.ShowBanner();
            }
            else
            {
                _logger.Log($"{nameof(RefreshBanner)} - asking {nameof(IAdsMediator)} to hide banner");
                Mediator.HideBanner();
            }
            _bannerActive.Value = needBanner;
        }

        private void DisposeCurrentMediator()
        {
            if (Mediator != null)
            {
                if (IsBannerActive.Value && Mediator.IsBannerVisible)
                {
                    Mediator.HideBanner();
                }
                Mediator.OnBannerLoaded -= BannerLoadedHandler;
                Mediator.OnBannerLoadFailed -= BannerLoadFailedHandler;
                Mediator.OnBannerClicked -= BannerClickedHandler;
            }
        }

        #endregion

        #region Reward

        public UniTask<AdLoadResult> LoadRewardVideoAsync(CancellationToken cancellationToken)
        {
            var timeout = (float)_settings.RewardLoadTimeout;
            return LoadRewardVideoAsync(timeout, cancellationToken);
        }

        public async UniTask<AdLoadResult> LoadRewardVideoAsync(float timeout, CancellationToken cancellationToken)
        {
            if (Mediator == null)
            {
                return AdLoadResult.NotReady();
            }
            if (timeout <= 0 && (!Mediator.IsInitialized || !Mediator.IsRewardedVideoReady))
            {
                return AdLoadResult.Timeout();
            }
            if (Mediator.IsInitialized && Mediator.IsRewardedVideoReady)
            {
                return AdLoadResult.Ready();
            }
            if (!_network.IsNetworkReachable)
            {
                _logger.LogWarning("Network is not reachable");
                return AdLoadResult.NoInternet();
            }

            return await LoadVideoAsync(LoadRewardVideoAsync, timeout, cancellationToken);
        }

        public async UniTask<AdShowResult> ShowRewardedVideoAsync(string placement, Action<IAdsImpression> clickedCallback)
        {
            var ready = Mediator.IsRewardedVideoReady;
            _logger.Log($"Show RewardVideo. {nameof(Mediator.IsRewardedVideoReady)} - {ready}");
            if (!ready)
            {
                return AdShowResult.NotReady();
            }

            var promise = new UniTaskCompletionSource<AdShowResult>();
            var rewardReceived = false;

            Mediator.OnRewardDisplayFailed += failed;
            Mediator.OnRewardHidden += hidden;
            Mediator.OnRewardClicked += clicked;
            Mediator.OnRewardReceived += rewarded;
            var result = AdShowResult.NotReady();
            try
            {
                _logger.Log($"Show RewardVideo for placement {placement}");
                _onBeforeShowRewardVideo.OnNext(Unit.Default);
                Mediator.ShowRewardVideo(placement);
                result = await promise.Task;
                _logger.Log($"Show RewardVideo finished with result {result}");
                return result;
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                if (result.State == AdShowState.Showed)
                {
                    _prefs.RewardedVideoCounter.Value++;
                }
                _onAfterShowRewardVideo.OnNext(Unit.Default);
                Mediator.OnRewardDisplayFailed -= failed;
                Mediator.OnRewardHidden -= hidden;
                Mediator.OnRewardClicked -= clicked;
                Mediator.OnRewardReceived -= rewarded;
            }

            void failed(string error)
            {
                promise.TrySetResult(AdShowResult.MediatorError(error));
            }

            void rewarded()
            {
                _logger.Log("RewardVideo - reward received");
                rewardReceived = true;
            }

            void clicked(IAdsImpression impression)
            {
                clickedCallback?.Invoke(impression);
            }

            void hidden(IAdsImpression impression)
            {
                promise.TrySetResult(rewardReceived ? AdShowResult.Showed(impression) : AdShowResult.UserCanceled());
            }
        }

        private async UniTask<AdLoadResult> LoadRewardVideoAsync()
        {
            _logger.Log("RewardVideo - Ask SDK to load reward video");
            if (Mediator.IsRewardedVideoReady)
            {
                return AdLoadResult.Ready();
            }
            var promise = new UniTaskCompletionSource<AdLoadResult>();
            Mediator.OnRewardLoaded += loaded;
            Mediator.OnRewardLoadFailed += failed;

            Mediator.LoadRewardVideo();
            try
            {
                return await promise.Task;
            }
            catch (Exception e)
            {
                return AdLoadResult.Unexpected(e.GetType().Name);
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                Mediator.OnRewardLoaded -= loaded;
                Mediator.OnRewardLoadFailed -= failed;
            }

            void loaded()
            {
                promise.TrySetResult(AdLoadResult.Ready());
            }

            void failed(string error)
            {
                promise.TrySetResult(AdLoadResult.MediatorError(error));
            }
        }

        #endregion

        #region Interstitial

        public async UniTask<AdLoadResult> LoadInterVideoAsync(float timeout, CancellationToken cancellationToken)
        {
            if (Mediator == null)
            {
                return AdLoadResult.NotReady();
            }
            if (timeout <= 0 && (!Mediator.IsInitialized || !Mediator.IsInterVideoReady))
            {
                return AdLoadResult.Timeout();
            }
            if (Mediator.IsInitialized && Mediator.IsRewardedVideoReady)
            {
                return AdLoadResult.Ready();
            }
            if (!_network.IsNetworkReachable)
            {
                _logger.LogWarning("Network is not reachable");
                return AdLoadResult.NoInternet();
            }
            return await LoadVideoAsync(LoadInterVideoAsync, timeout, cancellationToken);
        }

        public async UniTask<AdShowResult> ShowInterVideoAsync(string placement, Action<IAdsImpression> clickedCallback)
        {
            var ready = Mediator.IsInterVideoReady;
            _logger.Log($"Show InterVideo. {nameof(Mediator.IsInterVideoReady)} - {ready}");

            var promise = new UniTaskCompletionSource<AdShowResult>();

            Mediator.OnInterDisplayFailed += failed;
            Mediator.OnInterHidden += hidden;
            Mediator.OnInterClicked += clicked;
            var result = AdShowResult.NotReady();
            try
            {
                _logger.Log($"Show InterVideo");
                _onBeforeShowInterstitialVideo.OnNext(Unit.Default);
                Mediator.ShowInterVideo(placement);
                result = await promise.Task;
                _logger.Log($"Show InterVideo finished with result {result}");
                return result;
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                if (result.State == AdShowState.Showed)
                {
                    _prefs.InterstitialVideoCounter.Value++;
                }
                _onAfterShowInterstitialVideo.OnNext(Unit.Default);
                Mediator.OnInterDisplayFailed -= failed;
                Mediator.OnInterHidden -= hidden;
                Mediator.OnInterClicked -= clicked;
            }

            void failed(string error)
            {
                promise.TrySetResult(AdShowResult.MediatorError(error));
            }

            void hidden(IAdsImpression impression)
            {
                promise.TrySetResult(AdShowResult.Showed(impression));
            }

            void clicked(IAdsImpression impression)
            {
                clickedCallback?.Invoke(impression);
            }
        }

        private async UniTask<AdLoadResult> LoadInterVideoAsync()
        {
            _logger.Log("InterVideo - Ask SDK to load interstitial video");
            if (Mediator.IsInterVideoReady)
            {
                return AdLoadResult.Ready();
            }

            var promise = new UniTaskCompletionSource<AdLoadResult>();
            Mediator.OnInterLoaded += loaded;
            Mediator.OnInterLoadFailed += failed;

            Mediator.LoadInterVideo();
            try
            {
                return await promise.Task;
            }
            catch (Exception e)
            {
                return AdLoadResult.Unexpected(e.GetType().Name);
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                Mediator.OnInterLoaded -= loaded;
                Mediator.OnInterLoadFailed -= failed;
            }

            void loaded()
            {
                promise.TrySetResult(AdLoadResult.Ready());
            }

            void failed(string error)
            {
                promise.TrySetResult(AdLoadResult.MediatorError(error));
            }
        }

        #endregion

        public async UniTask<bool> WaitForInitialization(float timeout, CancellationToken cancellationToken)
        {
            var startTime = Time.time;
            while (Time.time - startTime < timeout)
            {
                if (Mediator.IsInitialized)
                {
                    return true;
                }
                await UniTask.Yield(cancellationToken);
            }
            return false;
        }

        private async UniTask<AdLoadResult> LoadVideoAsync(Func<UniTask<AdLoadResult>> repeater, float timeout, CancellationToken cancellationToken)
        {
            if (!Mediator.IsInitialized)
            {
                var time = Time.time;
                if (!await WaitForInitialization(timeout, cancellationToken))
                {
                    return AdLoadResult.Timeout();
                }
                timeout -= Time.time - time;
            }

            AdLoadResult result;
            var source = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            try
            {
                result = await repeater
                    .RepeatWhile(c => c.State == AdLoadState.Loaded, source.Token)
                    .Timeout(TimeSpan.FromSeconds(timeout));
            }
            catch (TimeoutException)
            {
                result = AdLoadResult.Timeout();
                source.Cancel();
            }
            catch (Exception e)
            {
                result = AdLoadResult.Unexpected(e.GetType().Name);
            }
            finally
            {
                await UniTask.SwitchToMainThread();
                source.Dispose();
            }
            return result;
        }
    }
}