using System;
using UniRx;
using UnityEngine;

namespace Magify
{
    internal class AnalyticsSessionTracker : IInitializable, IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Analytics);

        private readonly AnalyticsTracker _analyticsTracker;
        private readonly GeneralPrefs _generalPrefs;
        private readonly PlatformAPI _platform;
        private readonly CompositeDisposable _disposables = new();

        private IDisposable _intervalDisposable;

        private long SessionStarted
        {
            get => _generalPrefs.SessionStarted;
            set => _generalPrefs.SessionStarted = value;
        }

        private long LastActivity
        {
            get => _generalPrefs.LastActivity;
            set => _generalPrefs.LastActivity = value;
        }

        public AnalyticsSessionTracker(AnalyticsTracker analyticsTracker, GeneralPrefs generalPrefs, PlatformAPI platform)
        {
            _analyticsTracker = analyticsTracker;
            _generalPrefs = generalPrefs;
            _platform = platform;
        }

        void IInitializable.Initialize()
        {
            _generalPrefs.GlobalSessionCounter
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => TrackAppLaunch())
                .AddTo(_disposables);

            _analyticsTracker.OnEventSent
                .Subscribe(_ => TrackActivity())
                .AddTo(_disposables);

            _platform.OnBackground
                .Subscribe(_ => OnBackground())
                .AddTo(_disposables);

            _platform.OnForeground
                .Subscribe(_ => OnForeground())
                .AddTo(_disposables);

            CheckFailedSession();
            StartSessionActivityListening();
        }

        void IDisposable.Dispose()
        {
            _intervalDisposable?.Dispose();
            _intervalDisposable = null;
            _disposables.Dispose();
        }

        private void OnBackground()
        {
            TrackAppBackgroundEvent(SessionStarted, DateTime.UtcNow.ToUnixMilliseconds());
            _analyticsTracker.OnBackground();
        }

        private void OnForeground()
        {
            if (SessionStarted > 0)
            {
                // Unexpected SessionStarted value on foreground event. It should be zero at this point.
                // Possible problem: Magify Initialized in Awake method in first loaded scene (means before first OnApplicationFocus call)
                // Just do nothing in this case because session already started
                return;
            }
            _analyticsTracker.OnForeground();
            StartSessionActivityListening();
        }

        private void StartSessionActivityListening()
        {
            SessionStarted = DateTime.UtcNow.ToUnixMilliseconds();
            _intervalDisposable = Observable
                .Interval(TimeSpan.FromSeconds(30))
                .Subscribe(_ => TrackActivity());
            TrackActivity();
        }

        private void CheckFailedSession()
        {
            var sessionId = _generalPrefs.SessionId.Value;
            var openTimestamp = SessionStarted;
            var closeTimestamp = LastActivity;

            if (!string.IsNullOrEmpty(sessionId) && openTimestamp > 0 && closeTimestamp > 0)
            {
                TrackAppBackgroundEvent(openTimestamp, closeTimestamp);
            }
        }

        private void TrackAppBackgroundEvent(long start, long end)
        {
            _analyticsTracker.TrackAppBackgroundEvent(start, end);

            SessionStarted = 0;
            LastActivity = 0;
            _intervalDisposable?.Dispose();
            _intervalDisposable = null;
        }

        internal void TrackAppLaunch()
        {
            _analyticsTracker.TrackAppLaunch();
        }

        private void TrackActivity()
        {
            LastActivity = DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
        }
    }
}