using System;
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Magify.Model;
using UniRx;
using UnityEngine;

namespace Magify.Features
{
    internal class FeaturesProvider : IFeatures, IInitializable, IContextListener, IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Features);

        public event Action OnUpdate;

        public Dictionary<string, object> AllFeatures { get; private set; }
        public Dictionary<string, object> AllDefaultFeatures { get; private set; }

        private SortedDictionary<string, bool> _bufferBooleans = new();
        private SortedDictionary<string, double> _bufferNumbers = new();
        private SortedDictionary<string, string> _bufferStrings = new();

        private SortedDictionary<string, bool> _defaultBooleans = new();
        private SortedDictionary<string, double> _defaultNumbers = new();
        private SortedDictionary<string, string> _defaultStrings = new();

        private SortedDictionary<string, bool> _currentBooleans = new();
        private SortedDictionary<string, double> _currentNumbers = new();
        private SortedDictionary<string, string> _currentStrings = new();

        [CanBeNull]
        private Dictionary<string, FeatureConfig> _featureConfig;
        [CanBeNull]
        private Dictionary<string, FeatureConfig> _defaultFeatureConfig;

        private readonly GeneralPrefs _generalPrefs;
        private readonly CompositeDisposable _disposables = new();

        #region IFeatures implementaion

        public bool IsCurrentFeaturesParsed { get; private set; }

        public bool TryGetBool(string featureName, out bool result)
        {
            return TryGetFeature(featureName, _defaultBooleans, _currentBooleans, out result);
        }

        public bool GetBool(string featureName)
        {
            if (!TryGetBool(featureName, out var result))
            {
                throw new ArgumentException($"Can't find {nameof(Boolean)} feature {featureName}");
            }
            return result;
        }

        public bool TryGetNumber(string featureName, out double result)
        {
            return TryGetFeature(featureName, _defaultNumbers, _currentNumbers, out result);
        }

        public double GetNumber(string featureName)
        {
            if (!TryGetNumber(featureName, out var result))
            {
                throw new ArgumentException($"Can't find {nameof(Boolean)} feature {featureName}");
            }
            return result;
        }

        public bool TryGetString(string featureName, out string result)
        {
            return TryGetFeature(featureName, _defaultStrings, _currentStrings, out result);
        }

        public string GetString(string featureName)
        {
            if (!TryGetString(featureName, out var result))
            {
                throw new ArgumentException($"Can't find {nameof(Boolean)} feature {featureName}");
            }
            return result;
        }

        private static bool TryGetFeature<T>(string featureName, IReadOnlyDictionary<string, T> defaultMap, IReadOnlyDictionary<string, T> currentMap, out T result)
        {
            if (currentMap.TryGetValue(featureName, out result))
            {
                return true;
            }
            if (defaultMap.TryGetValue(featureName, out result))
            {
                return true;
            }
            result = default;
            return false;
        }

        #endregion

        public FeaturesProvider(GeneralPrefs generalPrefs)
        {
            _generalPrefs = generalPrefs;
        }

        void IInitializable.Initialize()
        {
            _generalPrefs.SubscriptionStatus
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => UpdateFeaturesOnTrigger())
                .AddTo(_disposables);

            _generalPrefs.InAppStatus
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => UpdateFeaturesOnTrigger())
                .AddTo(_disposables);

            _generalPrefs.AuthorizationStatus
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => UpdateFeaturesOnTrigger())
                .AddTo(_disposables);

            _generalPrefs.ReferrerId
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => UpdateFeaturesOnTrigger())
                .AddTo(_disposables);
        }

        void IDisposable.Dispose()
        {
            _disposables.Dispose();
        }

        public void OnProductPurchase()
        {
            UpdateFeaturesOnTrigger();
        }

        public void UpdateContext(CampaignsContext context, ContextKind kind)
        {
            if (kind is ContextKind.Default)
            {
                AllDefaultFeatures = BuildFeatureValueMap(context.Features).ToDictionary(c => c.Key, c => c.Value);
            }
            else
            {
                AllFeatures = BuildFeatureValueMap(context.Features).ToDictionary(c => c.Key, c => c.Value);
            }
            UpdateFeatures(context.Features, kind is ContextKind.Default);
        }

        public void ResetFeatures()
        {
            UpdateFeatures(default, true);
        }

        private void UpdateFeatures(Dictionary<string, FeatureConfig> features, bool isDefault)
        {
            bool isUpdated;
            if (isDefault)
            {
                _defaultFeatureConfig = features;
                isUpdated = UpdateDefaultFeatureValues(_defaultFeatureConfig);
            }
            else
            {
                _featureConfig = features;
                isUpdated = UpdateFeatureValues(_featureConfig);
                IsCurrentFeaturesParsed = _featureConfig != null;
            }

            if (isUpdated)
            {
                OnUpdate?.Invoke();
            }
        }

        private void UpdateFeaturesOnTrigger()
        {
            var isFeaturesUpdated = UpdateFeatureValues(_featureConfig);
            var isDefaultFeaturesUpdated = UpdateDefaultFeatureValues(_defaultFeatureConfig);

            if (isFeaturesUpdated || isDefaultFeaturesUpdated)
            {
                OnUpdate?.Invoke();
            }
        }

        private bool UpdateFeatureValues([CanBeNull] Dictionary<string, FeatureConfig> config)
        {
            return UpdateFeatures(config, ref _currentBooleans, ref _currentNumbers, ref _currentStrings);
        }

        private bool UpdateDefaultFeatureValues([CanBeNull] Dictionary<string, FeatureConfig> config)
        {
            return UpdateFeatures(config, ref _defaultBooleans, ref _defaultNumbers, ref _defaultStrings);
        }

        private IEnumerable<(string Key, object Value)> BuildFeatureValueMap(Dictionary<string, FeatureConfig> config)
        {
            foreach (var (key, featureConfig) in config)
            {
                var option = featureConfig.Options?.FirstOrDefault(IsSuitable);
                var rawValue = option?.Value ?? featureConfig.DefaultValue;
                if (rawValue is not null)
                {
                    yield return (key, rawValue);
                }
            }
        }

        private bool UpdateFeatures(
            [CanBeNull] Dictionary<string, FeatureConfig> config,
            ref SortedDictionary<string, bool> booleans,
            ref SortedDictionary<string, double> numbers,
            ref SortedDictionary<string, string> strings)
        {
            _bufferBooleans.Clear();
            _bufferNumbers.Clear();
            _bufferStrings.Clear();

            if (config != null)
            {
                foreach (var (key, rawValue) in BuildFeatureValueMap(config))
                {
                    ParseFeature(key, rawValue, _bufferBooleans, _bufferNumbers, _bufferStrings);
                }
            }

            if (!booleans.SequenceEqual(_bufferBooleans) || !numbers.SequenceEqual(_bufferNumbers) || !strings.SequenceEqual(_bufferStrings))
            {
                (booleans, _bufferBooleans) = (_bufferBooleans, booleans);
                (numbers, _bufferNumbers) = (_bufferNumbers, numbers);
                (strings, _bufferStrings) = (_bufferStrings, strings);
                return true;
            }

            return false;
        }

        private bool IsSuitable(FeatureOption option)
        {
            if (option?.Trigger is null)
            {
                return false;
            }

            if (option.Trigger.SubscriptionStatus?.IsCorresponded(_generalPrefs.SubscriptionStatus.Value) == false)
            {
                return false;
            }
            if (option.Trigger.InAppStatus?.IsCorresponded(_generalPrefs.InAppStatus.Value) == false)
            {
                return false;
            }
            if (option.Trigger.AuthorizationStatus?.IsCorresponded(_generalPrefs.AuthorizationStatus.Value) == false)
            {
                return false;
            }
            if (option.Trigger.PayingStatus?.IsCorresponded(_generalPrefs.SubscriptionStatus.Value, _generalPrefs.InAppStatus.Value) == false)
            {
                return false;
            }
            if (!option.Trigger.GetReferrerLimits().IsCorresponded(_generalPrefs.ReferrerId.Value))
            {
                return false;
            }
            if (!option.Trigger.GetPurchaseLimits().IsCorresponded(_generalPrefs.AllUsedProducts))
            {
                return false;
            }

            return true;
        }

        private static void ParseFeature(
            string key,
            object rawValue,
            in IDictionary<string, bool> booleans,
            in IDictionary<string, double> numbers,
            in IDictionary<string, string> strings)
        {
            _logger.Log($"Key={key}. Type={(rawValue != null ? rawValue.GetType().Name : "N/A")}. Value={rawValue}");
            switch (rawValue)
            {
                case bool value:
                    booleans[key] = value;
                    break;
                case short value:
                    numbers[key] = value;
                    break;
                case int value:
                    numbers[key] = value;
                    break;
                case long value:
                    numbers[key] = value;
                    break;
                case float value:
                    numbers[key] = value;
                    break;
                case double value:
                    numbers[key] = value;
                    break;
                case string value:
                    strings[key] = value;
                    break;
                default:
                    strings[key] = JsonFacade.SerializeObject(rawValue);
                    _logger.Log("Convert to json");
                    _logger.Log(strings[key]);
                    break;
            }
        }
    }
}