﻿using System;
using System.Buffers;
using System.Collections.Generic;
using System.IO;
using System.Text;
using UnityEngine;

namespace Magify
{
    public abstract class BinaryTypeHandler<T>
    {
        public virtual string TypeName { get; } = typeof(T).Name;
        public abstract int SizeOf(T value);
        public abstract void WriteTo(BinaryWriter writer, T value);
        public abstract T ReadFrom(BinaryReader reader);
    }

    public class BinaryEnumHandler<TEnum, TNumber> : BinaryTypeHandler<TEnum>
        where TEnum : unmanaged, Enum
        where TNumber : unmanaged
    {
        private readonly BinaryTypeHandler<TNumber> _numberType;

        public override string TypeName { get; }
        public override int SizeOf(TEnum value) => _numberType.SizeOf(ToNumber(value));

        public BinaryEnumHandler(BinaryTypeHandler<TNumber> numberType, bool useFullName)
        {
            _numberType = numberType;
            TypeName = useFullName ? typeof(TEnum).FullName : typeof(TEnum).Name;
        }

        public override void WriteTo(BinaryWriter writer, TEnum value)
        {
            var number = ToNumber(value);
            _numberType.WriteTo(writer, number);
        }

        public override TEnum ReadFrom(BinaryReader reader)
        {
            var number = _numberType.ReadFrom(reader);
            return FromNumber(number);
        }

        public static unsafe TNumber ToNumber(TEnum enumValue)
        {
            return *(TNumber*)(&enumValue);
        }

        public static unsafe TEnum FromNumber(TNumber numberValue)
        {
            return *(TEnum*)(&numberValue);
        }
    }

    internal class BinaryTypesPair<TType1, TType2> : BinaryTypeHandler<KeyValuePair<TType1, TType2>>
    {
        private readonly BinaryTypeHandler<TType1> _type1;
        private readonly BinaryTypeHandler<TType2> _type2;

        public override string TypeName { get; }

        public BinaryTypesPair(BinaryTypeHandler<TType1> t1, BinaryTypeHandler<TType2> t2)
        {
            _type1 = t1;
            _type2 = t2;
            TypeName = $"{typeof(TType1).Name}:{typeof(TType2).Name}";
        }

        public override int SizeOf(KeyValuePair<TType1, TType2> value)
        {
            return _type1.SizeOf(value.Key) + _type2.SizeOf(value.Value);
        }

        public override void WriteTo(BinaryWriter writer, KeyValuePair<TType1, TType2> value)
        {
            _type1.WriteTo(writer, value.Key);
            _type2.WriteTo(writer, value.Value);
        }

        public override KeyValuePair<TType1, TType2> ReadFrom(BinaryReader reader)
        {
            return new KeyValuePair<TType1, TType2>(_type1.ReadFrom(reader), _type2.ReadFrom(reader));
        }
    }

    internal class BinaryTypeBoolean : BinaryTypeHandler<bool>
    {
        public static BinaryTypeBoolean Shared { get; } = new();
        public override int SizeOf(bool _) => sizeof(bool);
        public override void WriteTo(BinaryWriter writer, bool value) => writer.Write(value);
        public override bool ReadFrom(BinaryReader reader) => reader.ReadBoolean();
    }

    internal class BinaryTypeChar : BinaryTypeHandler<char>
    {
        public static BinaryTypeChar Shared { get; } = new();
        public override int SizeOf(char _) => sizeof(char);
        public override void WriteTo(BinaryWriter writer, char value) => writer.Write(value);
        public override char ReadFrom(BinaryReader reader) => reader.ReadChar();
    }

    internal class BinaryTypeByte : BinaryTypeHandler<byte>
    {
        public static BinaryTypeByte Shared { get; } = new();
        public override int SizeOf(byte _) => sizeof(byte);
        public override void WriteTo(BinaryWriter writer, byte value) => writer.Write(value);
        public override byte ReadFrom(BinaryReader reader) => reader.ReadByte();
    }

    internal class BinaryTypeSByte : BinaryTypeHandler<sbyte>
    {
        public static BinaryTypeSByte Shared { get; } = new();
        public override int SizeOf(sbyte _) => sizeof(sbyte);
        public override void WriteTo(BinaryWriter writer, sbyte value) => writer.Write(value);
        public override sbyte ReadFrom(BinaryReader reader) => reader.ReadSByte();
    }

    internal class BinaryTypeInt16 : BinaryTypeHandler<short>
    {
        public static BinaryTypeInt16 Shared { get; } = new();
        public override int SizeOf(short _) => sizeof(short);
        public override void WriteTo(BinaryWriter writer, short value) => writer.Write(value);
        public override short ReadFrom(BinaryReader reader) => reader.ReadInt16();
    }

    internal class BinaryTypeUInt16 : BinaryTypeHandler<ushort>
    {
        public static BinaryTypeUInt16 Shared { get; } = new();
        public override int SizeOf(ushort _) => sizeof(short);
        public override void WriteTo(BinaryWriter writer, ushort value) => writer.Write(value);
        public override ushort ReadFrom(BinaryReader reader) => reader.ReadUInt16();
    }

    internal class BinaryTypeInt32 : BinaryTypeHandler<int>
    {
        public static BinaryTypeInt32 Shared { get; } = new();
        public override int SizeOf(int _) => sizeof(int);
        public override void WriteTo(BinaryWriter writer, int value) => writer.Write(value);
        public override int ReadFrom(BinaryReader reader) => reader.ReadInt32();
    }

    internal class BinaryTypeUInt32 : BinaryTypeHandler<uint>
    {
        public static BinaryTypeUInt32 Shared { get; } = new();
        public override int SizeOf(uint _) => sizeof(uint);
        public override void WriteTo(BinaryWriter writer, uint value) => writer.Write(value);
        public override uint ReadFrom(BinaryReader reader) => reader.ReadUInt32();
    }

    internal class BinaryTypeInt64 : BinaryTypeHandler<long>
    {
        public static BinaryTypeInt64 Shared { get; } = new();
        public override int SizeOf(long _) => sizeof(long);
        public override void WriteTo(BinaryWriter writer, long value) => writer.Write(value);
        public override long ReadFrom(BinaryReader reader) => reader.ReadInt64();
    }

    internal class BinaryTypeUInt64 : BinaryTypeHandler<ulong>
    {
        public static BinaryTypeUInt64 Shared { get; } = new();
        public override int SizeOf(ulong _) => sizeof(ulong);
        public override void WriteTo(BinaryWriter writer, ulong value) => writer.Write(value);
        public override ulong ReadFrom(BinaryReader reader) => reader.ReadUInt64();
    }

    internal class BinaryTypeSingle : BinaryTypeHandler<float>
    {
        public static BinaryTypeSingle Shared { get; } = new();
        public override int SizeOf(float _) => sizeof(float);
        public override void WriteTo(BinaryWriter writer, float value) => writer.Write(value);
        public override float ReadFrom(BinaryReader reader) => reader.ReadSingle();
    }

    internal class BinaryTypeDouble : BinaryTypeHandler<double>
    {
        public static BinaryTypeDouble Shared { get; } = new();
        public override int SizeOf(double _) => sizeof(double);
        public override void WriteTo(BinaryWriter writer, double value) => writer.Write(value);
        public override double ReadFrom(BinaryReader reader) => reader.ReadDouble();
    }

    internal class BinaryTypeDecimal : BinaryTypeHandler<decimal>
    {
        public static BinaryTypeDecimal Shared { get; } = new();
        public override int SizeOf(decimal _) => sizeof(decimal);
        public override void WriteTo(BinaryWriter writer, decimal value) => writer.Write(value);
        public override decimal ReadFrom(BinaryReader reader) => reader.ReadDecimal();
    }

    internal class BinaryTypeString : BinaryTypeHandler<string>
    {
        private readonly Encoding _encoding;

        public static BinaryTypeString Shared { get; } = new(Encoding.UTF8);

        public BinaryTypeString() : this(Encoding.UTF8)
        {
        }

        public BinaryTypeString(Encoding encoding)
        {
            _encoding = encoding;
        }

        public override int SizeOf(string value) => sizeof(int) + (value != null ? _encoding.GetByteCount(value) : 0);

        public override void WriteTo(BinaryWriter writer, string value)
        {
            if (value == null)
            {
                writer.Write(-1);
            }
            else if (value.Length == 0)
            {
                writer.Write(0);
            }
            else
            {
                var size = SizeOf(value);
                var buffer = ArrayPool<byte>.Shared.Rent(size);
                var bufferSize = _encoding.GetBytes(value, buffer);
                writer.Write(bufferSize);
                writer.Write(buffer, 0, bufferSize);
                ArrayPool<byte>.Shared.Return(buffer);
            }
        }

        public override string ReadFrom(BinaryReader reader)
        {
            var size = reader.ReadInt32();
            if (size == -1)
            {
                return null;
            }
            if (size == 0)
            {
                return string.Empty;
            }
            var buffer = ArrayPool<byte>.Shared.Rent(size);
            var read = reader.BaseStream.Read(buffer, 0, size);
            if (read != size)
            {
                ArrayPool<byte>.Shared.Return(buffer);
                return string.Empty;
            }
            var value = _encoding.GetString(buffer, 0, size);
            ArrayPool<byte>.Shared.Return(buffer);
            return value;
        }
    }

    internal class BinaryTypeDateTime : BinaryTypeHandler<DateTime>
    {
        public static BinaryTypeDateTime Shared { get; } = new();

        public override int SizeOf(DateTime _) => sizeof(long);

        public override void WriteTo(BinaryWriter writer, DateTime value)
        {
            writer.Write(value.ToBinary());
        }

        public override DateTime ReadFrom(BinaryReader reader)
        {
            var ticks = reader.ReadInt64();
            return DateTime.FromBinary(ticks);
        }
    }

    internal class BinaryTypeDateTimeLegacy : BinaryTypeHandler<DateTime>
    {
        public static BinaryTypeDateTimeLegacy Shared { get; } = new();

        public override int SizeOf(DateTime _) => sizeof(long);

        public override void WriteTo(BinaryWriter writer, DateTime value)
        {
            var unixMilliseconds = (value - DateTime.UnixEpoch).TotalMilliseconds;
            writer.Write(unixMilliseconds);
        }

        public override DateTime ReadFrom(BinaryReader reader)
        {
            var unixMilliseconds = reader.ReadDouble();
            return DateTime.UnixEpoch.AddMilliseconds(unixMilliseconds);
        }
    }

    internal class BinaryTypeVector2 : BinaryTypeHandler<Vector2>
    {
        public static BinaryTypeVector2 Shared { get; } = new();
        public override int SizeOf(Vector2 _) => sizeof(float) * 2;

        public override void WriteTo(BinaryWriter writer, Vector2 value)
        {
            writer.Write(value.x);
            writer.Write(value.y);
        }

        public override Vector2 ReadFrom(BinaryReader reader)
        {
            var x = reader.ReadSingle();
            var y = reader.ReadSingle();
            return new Vector2(x, y);
        }
    }

    internal class BinaryTypeVector3 : BinaryTypeHandler<Vector3>
    {
        public static BinaryTypeVector3 Shared { get; } = new();
        public override int SizeOf(Vector3 _) => sizeof(float) * 3;

        public override void WriteTo(BinaryWriter writer, Vector3 value)
        {
            writer.Write(value.x);
            writer.Write(value.y);
            writer.Write(value.z);
        }

        public override Vector3 ReadFrom(BinaryReader reader)
        {
            var x = reader.ReadSingle();
            var y = reader.ReadSingle();
            var z = reader.ReadSingle();
            return new Vector3(x, y, z);
        }
    }

    internal class BinaryTypeNullable<T> : BinaryTypeHandler<T?>
        where T: struct
    {
        public static BinaryTypeNullable<T> Shared { get; } = new();

        public override int SizeOf(T? value)
        {
            return
                sizeof(bool) // there is value or not
              + BinaryTypeHandlerUtils.HandlerForPrimitive<T>().SizeOf(value ?? default); // value size
        }

        public override void WriteTo(BinaryWriter writer, T? value)
        {
            var handler = BinaryTypeHandlerUtils.HandlerForPrimitive<T>();
            writer.Write(value.HasValue);
            handler.WriteTo(writer, value ?? default);
        }

        public override T? ReadFrom(BinaryReader reader)
        {
            var handler = BinaryTypeHandlerUtils.HandlerForPrimitive<T>();
            var hasValue = reader.ReadBoolean();
            var value = handler.ReadFrom(reader);
            return hasValue ? value : null;
        }
    }
}